"""
Region Object
"""

###########
# Imports #
# #########
from .utils import validate_list_of_strings
from .interface import send_request
from .constants import REGIONS_URL

class Region:
    """Region Object Constructor

        Attributes:
            name (str): The display name of the region. This will be a
                full name that is used in the control panel and other interfaces.
            slug (str): A human-readable string that is used as a unique
                identifier for each region.
            features (:type: list of str): This attribute is set to an array
                which contains features available in this region
            available (bool): This is a boolean value that represents whether
                new Droplets can be created in this region.
            sizes (:type: list of str): This attribute is set to an array
                which contains the identifying slugs for the sizes available in this region.

        Rasies:
            ValueError - Invalid arguments
    """
    def __init__(self, name, slug, features, available, sizes):

        if not isinstance(name, str):
            raise ValueError("Name must be a string")

        self.__name = name

        if not isinstance(slug, str):
            raise ValueError("Slug must be a string")

        self.__slug = slug

        try:
            validate_list_of_strings(features)
        except ValueError as value_error:
            raise ValueError("Features must be a list of strings") from value_error

        self.__features = features

        if not isinstance(available, bool):
            raise ValueError("Available must be a boolean")

        self.__available = available

        try:
            validate_list_of_strings(sizes)
        except ValueError as value_error:
            raise ValueError("Sizes must be a list of strings") from value_error
        self.__sizes = sizes

    def name(self):
        """Region Name Accessor

            Returns:
                name (str): Region Name
        """
        return self.__name

    def slug(self):
        """Region Slug Accessor

            Returns:
                slug (str): Region Slug
        """
        return self.__slug

    def features(self):
        """Region Features Accessor

            Returns:
                features (:type: list of str): Region features
        """
        return self.__features

    def available(self):
        """Region Availble Accessor

            Returns:
                available (bool): Region Availability
        """
        return self.__available

    def sizes(self):
        """Region Sizes Accessor

            Returns:
                sizes (:type: list of str): Region sizes
        """
        return self.__sizes

    @classmethod
    def list_regions(cls):
        """List Regions

            Returns:
                regions (:type: obj Region): List of regions

            Raises:
                ValueError - Failed to create region object
                NotAuthorized - Request was not authorized.
                TooManyRequests - Too many requests.
                InternalError - Internal error.
        """
        ret = send_request("GET", REGIONS_URL)

        regions = []
        for reg in ret['regions']:
            regions.append(Region(**reg))

        return regions



"""
    Digital Ocean Image Module
"""
from .constants import VALID_DIST_TYPES, VALID_IMAGE_TYPES, VALID_IMAGE_STATUS
from .utils import validate_list_of_strings, validate_date_time

class Image:
    """Digital Ocean Image Object

        Args:
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            name (str): The display name that has been given to an image.
            distribution (str): The name of a custom image's distribution.
            public (bool): This is a boolean value that indicates whether
                the image in question is public or not.
            regions (:obj:list of str): This attribute is an array of the regions that
                the image is available in. The regions are represented by their
                identifying slug values.
            id (int, optional): A unique number that can be used to identify and reference
            a specific image.
            type (int, optional): Describes the kind of image. It may be one of "snapshot",
                "backup", or "custom". This specifies whether an image is a user-generated
                Droplet snapshot, automatically created Droplet backup, or a user-provided
                virtual machine image.
            created_at (str, optional): A time value given in ISO8601 combined date and time
                format that represents when the image was created.
            slug (str, optional): A uniquely identifying string that is associated with each of the
                DigitalOcean-provided public images. These can be used to reference a public
                image as an alternative to the numeric id
            created_at (str, optional): A time value given in ISO8601 combined date and time format
                that represents when the image was created.
            min_disk_size (str, optional): The minimum disk size in GB required for a
                Droplet to use this image.
            size_gigabyes (float, optional): The size of the image in gigabytes.
            description (str, optional): An optional free-form text field to describe an image.
            tags (:type: list of str, optional): A flat array of tag names as strings to be applied
                to the resource.  Tag names may be for either existing or new tags.
            status (:type: list of str, optional): A status string indicating
                the state of a custom image. This may be NEW, available,
                pending, deleted, or retired.

        Raises:
            ValueError - Invalid arguments
    """
    def __init__(self, **kwargs):

        #TODO: Make common validation functions

        #
        # Required Args
        try:
            if not isinstance(kwargs['name'], str):
                raise ValueError("Name must be a string")
            self.__name = kwargs['name']
        except KeyError as key_error:
            raise ValueError("Name is a required argument") from key_error

        try:
            if kwargs['distribution'] not in VALID_DIST_TYPES:
                raise ValueError(f"Image distribution must be one of: {VALID_DIST_TYPES}")
            self.__distribution = kwargs['distribution']
        except KeyError as key_error:
            raise ValueError("Distribution is a required argument") from key_error

        try:
            if not isinstance(kwargs['public'], bool):
                raise ValueError("Public must be a boolean")
            self.__public = kwargs['public']
        except KeyError as key_error:
            raise ValueError("Public is a required argument") from key_error

        try:
            validate_list_of_strings(kwargs['regions'])
        except KeyError as key_error:
            raise ValueError("Region is a required argument") from key_error
        except ValueError as value_error:
            raise ValueError("Regions must be a list of strings") from value_error
        self.__regions = kwargs['regions']


        #
        # Optional Args
        if 'id' in kwargs:
            if not isinstance(kwargs['id'], int):
                raise ValueError("ID must be an integer")
            self.__id = kwargs['id']
        else:
            self.__id = None

        if 'type' in kwargs:
            if kwargs['type'] not in VALID_IMAGE_TYPES:
                raise ValueError(f"Invalid Image Type: {kwargs['type']}")
            self.__type = kwargs['type']
        else:
            self.__type = None

        if 'slug' in kwargs:
            if not isinstance(kwargs['slug'], str):
                raise ValueError(f"Image Slug must be a string: {kwargs['slug']}")
            self.__slug = kwargs['slug']
        else:
            self.__slug = None

        if 'created_at' in kwargs:
            if not isinstance(kwargs['created_at'], str):
                raise ValueError("created_at must be a string")

            try:
                validate_date_time(kwargs['created_at'])
            except ValueError as value_error:
                raise ValueError("Date must be an ISO8601 formated string") from value_error
            self.__created_at = kwargs['created_at']
        else:
            self.__created_at = None

        if 'min_disk_size' in kwargs:
            if not isinstance(kwargs['min_disk_size'], int) or kwargs['min_disk_size'] < 0:
                raise ValueError("Min Disk Size must be an integer >= 0")
            self.__min_disk_size = kwargs['min_disk_size']
        else:
            self.__min_disk_size = None

        if 'size_gigabytes' in kwargs:
            if not isinstance(kwargs['size_gigabytes'], float):
                raise ValueError("Image gigabyte size must be a float")
            self.__size_gigabytes = kwargs['size_gigabytes']
        else:
            self.__size_gigabytes = None

        if 'description' in kwargs:
            if not isinstance(kwargs['description'], str):
                raise ValueError("Description must be a string")
            self.__description = kwargs['description']
        else:
            self.__description = None

        if 'tags' in kwargs:
            try:
                validate_list_of_strings(kwargs['tags'])
            except ValueError as value_error:
                raise ValueError("Tags must be a list of strings") from value_error
            self.__tags = kwargs['tags']
        else:
            self.__tags = []

        if 'status' in kwargs:
            if kwargs['status'] not in VALID_IMAGE_STATUS:
                raise ValueError(f"Status must be one of the following: {VALID_IMAGE_STATUS}")
            self.__status = kwargs['status']
        else:
            self.__status = None

    def id(self):
        """Image ID Accessor

            Returns:
                id (int): Image ID.
        """
        return self.__id

    def name(self):
        """Image Name Accessor

            Returns:
                name (str): Image Name
        """
        return self.__name

    def type(self):
        """Image Type Accessor

            Returns:
                type (str): Image type
        """
        return self.__type

    def distribution(self):
        """Image Distribution Accessor

            Returns:
                distribution (str): Image distribution
        """
        return self.__distribution

    def slug(self):
        """Image Slug Accessor

            Returns:
                slug (str): Image slug
        """
        return self.__slug

    def public(self):
        """Image Public Accessor

            Returns:
                public (bool): Image public status
        """
        return self.__public

    def regions(self):
        """Image Regions Accessor

            Returns:
                regions (:type: list of str): List of regions the image is available in.
        """
        return self.__regions

    def created_at(self):
        """Image Created at Accessor

            Returns:
                created_at (str): Image created at time
        """
        return self.__created_at

    def min_disk_size(self):
        """Image Min Disk Size Accessor

            Returns:
                min_disk_size (int): Image minimum disk size
        """
        return self.__min_disk_size

    def size_gigabytes(self):
        """Image Size in gigabytes Accessor

            Returns:
                size_gigabytes (float): Image size in gigabytes
        """
        return self.__size_gigabytes

    def description(self):
        """Image description Accessor

            Returns:
                description (str): Image description
        """
        return self.__description

    def tags(self):
        """Image Tags Accessor

            Returns:
                tags (:type: list of str): List of Image tags.
        """
        return self.__tags

    def status(self):
        """Image Status Accessor

            Returns:
                status (str): Image status.
        """
        return self.__status



"""Cloud Provider Network Interface"""

###########
# Imports #
###########
import os
import json
import requests

from .exceptions import NotAuthorized, NotFound, InternalError, InvalidRequest
from .constants import(BASE_URL, HTTP_OK, HTTP_CREATED, HTTP_DELETE,
    HTTP_NOT_AUTHORIZED, HTTP_NOT_FOUND, HTTP_TOO_MANY_REQUESTS,
    HTTP_INTERNAL_ERROR, HTTP_UNPROCESSABLE_ENTITY)

def send_request(action, url, data=None):
    """Send HTTP(S) Request and process return.

        Args:
            action (str) - HTTP action.
            url (str) - URL to send the request to.
            data (dict, optional) - Arguments to pass in the body of the request.

        Return:
            json (dict) - Returned data in a JSON format.

        Raises:
            ValueError - Invalid HTTP action
            NotAuthorized - Request not authorized
            NotFound - Object not found
            TooManyRequests - Rate limit reached
            InternalError - Internal server error
            InvalidRequest - Invalid request.
    """

    #
    # Read token from the env
    token = os.getenv('DO_TOKEN')

    headers = {
        'Content-type': 'application/json',
        'Authorization': f"Bearer {token}"
    }

    url = f"{BASE_URL}{url}"

    if action == 'GET':
        ret = requests.get(url, headers=headers)
        #XXX need to handle pagination
        if ret.status_code != HTTP_OK:
            process_error(ret.status_code, ret.json()['message'])
        return ret.json()

    elif action == 'POST':
        ret = requests.post(url, data=json.dumps(data), headers=headers)
        if ret.status_code != HTTP_CREATED:
            process_error(ret.status_code, ret.json()['message'])
        return ret.json()

    elif action == 'PUT':
        #XXX
        pass

    elif action == 'PATCH':
        #XXX
        pass

    elif action == 'DELETE':
        ret = requests.delete(url, headers=headers)
        if ret.status_code != HTTP_DELETE:
            process_error(ret.status_code, ret.json()['message'])
        return {}

    else:
        raise ValueError("Unexpected REST Action")

def process_error(status_code, msg):
    """Process Error when a non successful status code is returned

        Args:
            status_code (init): HTTP Status Code
            msg (str): Error message returned by the server

        Raises:
            NotAuthorized - Request not authorized
            NotFound - Object not found
            TooManyRequests - Rate limit reached
            InternalError - Internal server error
            InvalidRequest - Invalid request.
    """
    if status_code == HTTP_NOT_AUTHORIZED:
        raise NotAuthorized
    elif status_code == HTTP_NOT_FOUND:
        raise NotFound
    elif status_code == HTTP_TOO_MANY_REQUESTS:
        pass #XXX fill me out
    elif status_code == HTTP_INTERNAL_ERROR:
        print(msg)#XXX need to add msg
        raise InternalError
    elif status_code == HTTP_UNPROCESSABLE_ENTITY:
        raise InvalidRequest(msg)
    else:
        pass



"""
Networks Object
"""
from ..constants import *
from ..utils import *
from .v4 import V4
from .v6 import V6

class Networks:
    """Networks Object Constructor

        Args:
            v4 (:type: list of obj V4): List of V4 network objects
            v6 (:type: list of obj V6): List of V6 network objects


        Attributes:
            v4 (:type: list of obj V4): List of V4 network objects
            v6 (:type: list of obj V6): List of V6 network objects

        Raises:
            ValueError - Invalid Argument
    """
    def __init__(self, v4, v6):

        self.__v4 = []
        for net in v4:
            try:
                self.__v4.append(V4(**net))
            except ValueError as value_error:
                raise ValueError(
                    f"Failed to create V4 Network object: {value_error}"
                ) from value_error

        self.__v6 = []
        for net in v6:
            try:
                self.__v6.append(V6(**net))
            except ValueError as value_error:
                raise ValueError(
                    f"Failed to create V6 Network object: {value_error}"
                ) from value_error



    def v4(self):
        """Networks v4 Accessor

            Returns:
                v4 (:type: list of obj v4): List of IPv4 networks.
        """
        return self.__v4

    def v6(self):
        """Networks v6 Accessor

            Returns:
                v6 (:type: list of obj v6): List of IPv6 networks.
        """
        return self.__v6



"""
Digital Ocean Droplet Modeule
"""

import re

from .constants import DROPLET_URL, VALID_DOPLET_STATUS
from .interface import send_request
from .utils import validate_list_of_strings, validate_list_of_ints, validate_date_time
from .window import Window
from .image import Image
from .size import Size
from .networks import Networks
from .region import Region

class Droplet:
    """Droplet Object
        Args:
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            id (int): A unique identifier for each Droplet instance.
                This is automatically generated upon Droplet creatio
            name (string): The human-readable name set for the Droplet instance.
            memory (int): Memory of the Droplet in megabytes.
            vcpus (int): The number of virtual CPUs.
            disk (int): The size of the Droplet's disk in gigabytes.
            locked (bool): A boolean value indicating whether the Droplet has been locked,
                preventing actions by users.
            status (string): A status string indicating the state of the Droplet instance.
                This may be "new", "active", "off", or "archive".
            created_at (string): A time value given in ISO8601 combined date and time format that
                represents when the Droplet was created.
            features (:type: list of str): An array of features enabled on this Droplet.
            backup_ids (:type: list of int): An array of backup IDs of any backups that have been
                taken of the Droplet instance. Droplet backups are enabled at the time
                of the instance creation.
            next_backup_window(:type: obj Window): The details of the Droplet's backups feature,
                if backups are configured for the Droplet. This object contains keys for
                the start and end times of the window during which the backup will start.
            snapshot_ids (:type: list of int): An array of snapshot IDs of any snapshots created
                from the Droplet instance.
            image (:type: obj Image): Image used to build droplet.
            volume_ids(:type: list of str): A flat array including the unique identifier
                for each Block Storage volume attached to the Droplet.
            size (:type: obj Size): Class of Droplets created from this size.
            size_slug (str): The unique slug identifier for the size of this Droplet.
            networks (:type: obj Networks): The details of the network that are configured for
                the Droplet instance. This is an object that contains keys for IPv4 and IPv6.
                The value of each of these is an array that contains objects describing
                an individual IP resource allocated to the Droplet. These will define attributes
                like the IP address, netmask, and gateway of the specific network depending on the
                type of network it is.
            region (:type: obj Region): Region the droplet resides.
            tags (:type: list of str): An array of Tags the Droplet has been tagged with.
            vpc_uuid (str, optional): A string specifying the UUID of the VPC to which the
            Droplet is assigned.

        Raises:
            ValueError: Invalid Arguments
    """
    def __init__(self, **kwargs):
        #
        # Required Args
        try:
            if not isinstance(kwargs['id'], int):
                raise ValueError(f"id must be an int: {kwargs['id']}")

            self.__id = kwargs['id']
        except KeyError as key_error:
            raise ValueError("ID is a required argument") from key_error

        try:
            if not isinstance(kwargs['name'], str):
                raise ValueError(f"Name must be a string: {kwargs['name']}")
            self.__name = kwargs['name']
        except KeyError as key_error:
            raise ValueError("Name is a required argument") from key_error

        try:
            if not isinstance(kwargs['memory'], int):
                raise ValueError(f"Memory must be an integer: {kwargs['memory']}")

            self.__memory = kwargs['memory']
        except KeyError as key_error:
            raise ValueError("Memory is a required argument") from key_error


        try:
            if not isinstance(kwargs['vcpus'], int):
                raise ValueError(f"vcpus must be an int: {kwargs['vcpus']}")

            self.__vcpus = kwargs['vcpus']
        except KeyError as key_error:
            raise ValueError("VCPUS is a required argument") from key_error


        try:
            if not isinstance(kwargs['disk'], int):
                raise ValueError(f"disk must be an int: {kwargs['disk']}")

            self.__disk = kwargs['disk']
        except KeyError as key_error:
            raise ValueError("Disk is a required argument") from key_error

        try:
            if not isinstance(kwargs['locked'], bool):
                raise ValueError(f"locked must be a bool: {kwargs['locked']}")

            self.__locked = kwargs['locked']
        except KeyError as key_error:
            raise ValueError("Locked is a required argument") from key_error


        try:
            if kwargs['status'] not in VALID_DOPLET_STATUS:
                raise ValueError(f"status not valid: {kwargs['status']}")

            self.__status = kwargs['status']
        except KeyError as key_error:
            raise ValueError("Status is a required argument") from key_error

        try:
            if not isinstance(kwargs['created_at'], str):
                raise ValueError(f"created_at must be a string: {kwargs['created_at']}")
            
            validate_date_time(kwargs['created_at'])

            self.__created_at = kwargs['created_at']
        except ValueError as value_error:
            raise ValueError("Date must be an ISO8601 formated string") from value_error
        except KeyError as key_error:
            raise ValueError("Created_at is a required argument") from key_error


        try:
            validate_list_of_strings(kwargs['features'])
            self.__features = kwargs['features']
        except ValueError as value_error:
            raise ValueError(
                f"features must be a list of strings: {kwargs['featurs']}") from value_error
        except KeyError as key_error:
            raise ValueError("Features is a required argument") from key_error

        try:
            validate_list_of_ints(kwargs['backup_ids'])
            self.__backup_ids = kwargs['backup_ids']
        except ValueError as value_error:
            raise ValueError(
                f"backup_ids must be a list of ints: {kwargs['backup_ids']}") from value_error
        except KeyError as key_error:
            raise ValueError("backup_ids is a required argument") from key_error

        try:
            if kwargs['next_backup_window'] is not None:
                window = Window(**kwargs['next_backup_window'])
                self.__next_backup_window = window
            else:
                self.__next_backup_window = None

        except ValueError as value_error:
            raise ValueError(f"Failed to create window object: {value_error}") from value_error
        except KeyError as key_error:
            raise ValueError("next_backup_window is a required argument") from key_error


        try:
            validate_list_of_ints(kwargs['snapshot_ids'])

            self.__snapshot_ids = kwargs['snapshot_ids']
        except ValueError as value_error:
            raise ValueError(
                f"snapshot_ids must be a list of ints: {kwargs['snapshot_ids']}") from value_error
        except KeyError as key_error:
            raise ValueError("snapshot_ids is a required argument") from key_error

        try:
            img = Image(**kwargs['image'])

            self.__image = img
        except ValueError as value_error:
            raise ValueError(f"Failed to create Image object: {value_error}") from value_error
        except KeyError as key_error:
            raise ValueError("Image is a required argument") from key_error


        try:
            validate_list_of_strings(kwargs['volume_ids'])

            self.__volume_ids = kwargs['volume_ids']
        except ValueError as value_error:
            raise ValueError("volume_ids must be a list of strings") from value_error
        except KeyError as key_error:
            raise ValueError("volume_ids is a required argument") from key_error

        try:
            self.__size = Size(**kwargs['size'])
        except ValueError as value_error:
            raise ValueError(f"Failed to create size object: {value_error}") from value_error
        except KeyError as key_error:
            raise ValueError("size is a required argument") from key_error

        try:
            if not isinstance(kwargs['size_slug'], str):
                raise ValueError(f"size_slug must be a str: {kwargs['size_slug']}")
            self.__size_slug = kwargs['size_slug']
        except KeyError as key_error:
            raise ValueError("size_slug is a requried argument") from key_error

        try:
            self.__networks = Networks(**kwargs['networks'])
        except ValueError as value_error:
            raise ValueError(f"Failed to create networks object: {value_error}") from value_error
        except KeyError as key_error:
            raise ValueError("networks is a required argument") from key_error

        try:
            self.__region = Region(**kwargs['region'])
        except ValueError as value_error:
            raise ValueError(f"Failed to create a region object: {value_error}") from value_error
        except KeyError as key_error:
            raise ValueError("region is a required object") from key_error

        try:
            validate_list_of_strings(kwargs['tags'])

            self.__tags = kwargs['tags']
        except ValueError as value_error:
            raise ValueError("tags must be a list of strings") from value_error
        except KeyError as key_error:
            raise ValueError("tags is a required argument") from key_error

        #
        # Optional
        if 'vpc_uuid' in kwargs:
            if not isinstance(kwargs['vpc_uuid'], str):
                raise ValueError("vpc_uusid must be a string")

            self.__vpc_uuid = kwargs['vpc_uuid']
        else:
            self.__vpc_uuid = None

    def id(self):
        """Droplet ID accessor

        Returns:
            id (int): Droplet ID
        """

        return self.__id

    def name(self):
        """Droplet Name Accessor

        Returns:
            name (str): Droplet name
        """
        return self.__name

    def memory(self):
        """Droplet Memory Accessor

        Returns:
            memory (int): Droplet Memory
        """
        return self.__memory

    def vcpus(self):
        """Droplet VCPUs Accessor

        Returns:
            vcpus (int): Droplet VCPUs
        """
        return self.__vcpus

    def disk(self):
        """Droplet Disk Accessor

        Returns:
            disk (int): Droplet Disks
        """
        return self.__disk

    def locked(self):
        """Droplet Locked Accessor

        Returns:
            locked (bool): Droplet Locked status
        """
        return self.__locked

    def status(self):
        """Droplet Status Accessor

        Returns:
            status (str): Droplet status
        """
        return self.__status

    def created_at(self):
        """Droplet Created At Accessor

        Returns:
            created_at (str): Droplet Created At time
        """
        return self.__created_at

    def features(self):
        """Droplet Features Accessor

        Returns:
            features (:type: list of str): List of droplet features.
        """
        return self.__features

    def next_backup_window(self):
        """Droplet Next Backup Window Accessor

        Returns:
            next_backup_window (:type: obj Window): Next backup window
        """
        return self.__next_backup_window

    def snapshot_ids(self):
        """Droplet Snapshot IDs Accessors

        Returns:
            snapshot_ids (:type: list of int): List of snapshot  IDs.
        """
        return self.__snapshot_ids

    def image(self):
        """Droplet Image Accessor

        Returns:
            image (:type: obj Image): Droplet Image
        """
        return self.__image

    def volume_ids(self):
        """Droplet Volume IDs Accessor

        Returns:
            volume_ids (:type: list of str): List of droplet volume Ids
        """
        return self.__volume_ids

    def size(self):
        """Droplet Size Accessor

        Returns:
            size (:type: obj Size): Droplet size information
        """
        return self.__size

    def size_slug(self):
        """Droplet Size Slug Accessor

        Returns:
            size_slug (str): Droplet Size Slug
        """
        return self.__size_slug

    def networks(self):
        """Droplet Networks Accessor

        Returns:
            networks (:type: obj Networks): Droplet Networks
        """
        return self.__networks

    def region(self):
        """Droplet Region Accessor

        Returns:
            regions (:type: obj Region): The region the droplet resides
        """
        return self.__region

    def tags(self):
        """Droplet Tags Accessor

        Returns:
            tags (:type: list of str): List of droplet tags
        """
        return self.__tags

    def vpc_uuid(self):
        """Droplet VPC UUID Accessor

        Returns:
            vpc_uuid (str): Droplet VPC UUID
        """
        return self.__vpc_uuid

    def backup_ids(self):
        """Droplet Backup IDs Accessor

        Returns:
            backup_ids (:type: list of int): List of backup ids
        """
        return self.__backup_ids

    @classmethod
    def list_droplets(cls):
        """List droplets

        Returns:
            droplets (:type: obj Droplet): List of droplet objects

        Raises:
            NotAuthorized - Request was not authorized.
            TooManyRequests - Too many requests.
            InternalError - Internal error.
        """
        ret = send_request("GET", DROPLET_URL)

        droplets = []
        for drop in ret['droplets']:
            droplets.append(Droplet(**drop))

        return droplets

    @classmethod
    def create_droplet(cls, data):
        """Create single droplet.

        Args:
            data (dict) - Dictionary representation of droplet parameters.

        Returns:
            droplet (:type: obj Droplet): Created droplet.

        Raises:
            NotAuthorized - Request was not authorized.
            TooManyRequests - Too many requests.
            InternalError - Internal error.
        """
        ret = send_request("POST", DROPLET_URL, data)

        droplet = Droplet(**ret['droplet'])

        return droplet

    @classmethod
    def get_droplet(cls, id):
        """Retrieve Droplet.

        Args:
            id (int) - Droplet ID

        Returns:
            droplet (:type: obj Droplet) - Droplet object

        Raises:
            NotAuthorized - Request was not authorized.
            NotFound - Droplet does not exist.
            TooManyRequests - Too many requests.
            InternalError - Internal error.
        """

        ret = send_request("GET", f"{DROPLET_URL}/{id}")

        droplet = Droplet(**ret['droplet'])

        return droplet

    def delete(self):
        """Delete an existing droplet.

        Returns:
            True - Droplet was deleted

        Raises:
            NotAuthorized - Request was not authorized.
            NotFound - Droplet does not exist.
            TooManyRequests - Too many requests.
            InternalError - Internal error.
        """
        send_request("DELETE", f"{DROPLET_URL}/{self.__id}")

        return True



"""
Window Object
"""
class Window:
    """Window Object Constructor

        Attributes:
            start (str): A time value given in ISO8601 combined date and time
                format specifying the start of the Droplet's backup window.
            end (str): A time value given in ISO8601 combined date and time
                format specifying the end of the Droplet's backup window.

        Raises:
            ValueError: Invalid arguments
    """
    def __init__(self, start, end):
        #XXX better validation
        if not isinstance(start, str):
            raise ValueError("Start must be a string")

        self.__start = start

        if not isinstance(end, str):
            raise ValueError("End must be a string")

        self.__end = end

    def start(self):
        """Window Start Accessor

            Returns:
                start (str): Backup window start time
        """
        return self.__start

    def end(self):
        """Window End Accessor

            Returns:
                end (str): Backup window wend time
        """
        return self.__end



"""Custom Exceptions"""

class NotAuthorized(Exception):
    """Not Authorised Exception"""
    def __init__(self):
        super().__init__('Failed to authenticate with Digital Ocean')

class NotFound(Exception):
    """Not Found Exception"""
    def __init__(self):
        super().__init__("Object not found")

class InternalError(Exception):
    """Internal Error Exception"""
    def __init__(self):
        super().__init__("Internal Error")

class InvalidRequest(Exception):
    """Invalid Request"""
    def __init__(self, msg):
        super().__init__(f"Invalid Request: {msg}")



"""
HTTP Status Codes
"""
HTTP_OK = 200
HTTP_CREATED = 202
HTTP_DELETE = 204
HTTP_NOT_AUTHORIZED = 401
HTTP_NOT_FOUND = 404
HTTP_UNPROCESSABLE_ENTITY = 422
HTTP_TOO_MANY_REQUESTS = 429
HTTP_INTERNAL_ERROR = 500

"""
Digital Ocean Resource Endpoints
"""
BASE_URL = 'https://api.digitalocean.com/v2'
DROPLET_URL = "/droplets"
REGIONS_URL = "/regions"
SIZES_URL = "/sizes"

"""
Image Constants
"""
VALID_IMAGE_TYPES = ['base', 'snapshot', 'backup', 'custom','admin']
VALID_DIST_TYPES = ['Arch Linux', 'CentOS', 'CoreOS', 'Debian', 'Fedora',
    'Fedora Atomic', 'FreeBSD', 'Gentoo', 'openSUSE', 'RancherOS', 'Ubuntu', 'Unknown']
VALID_IMAGE_STATUS = ['NEW', 'new', 'available', 'pending','deleted','retired']

"""
Droplet Constants
"""
VALID_DOPLET_STATUS = ["new", "active", "off", "archive"]



"""
Size Object
"""

from .utils import validate_list_of_strings
from .interface import send_request
from .constants import SIZES_URL

class Size:
    """Size Object Constructor

        Args:
            **kwargs: Arbitrary keyword arguments.

        Attributes:
            slug (str): A human-readable string that is used to uniquely
                identify each size.
            memory (int): The amount of RAM allocated to Droplets created
                of this size. The value is represented in megabytes.
            vcpus (int): The integer of number CPUs allocated to Droplets
                of this size.
            disk (int): The amount of disk space set aside for Droplets of
                this size. The value is represented in gigabytes.
            transfer (float): The amount of transfer bandwidth that is
                available for Droplets created in this size. This only counts
                traffic on the public interface. The value is given in
                terabytes.
            price_monthly (float): This attribute describes the monthly cost
                of this Droplet size if the Droplet is kept for an entire month.
                The value is measured in US dollars.
            price_hourly (float): This describes the price of the Droplet size
                as measured hourly. The value is measured in US dollars.
            regions (:type: list of str): An array containing the region slugs
                where this size is available for Droplet creates.
            available (bool): This is a boolean value that represents whether
                new Droplets can be created with this size.
            description (str): A string describing the class of Droplets
                created from this size. For example: Basic, General Purpose,
                CPU-Optimized, Memory-Optimized, or Storage-Optimized.

        Raises:
            ValueError - Invalid arguments
    """
    def __init__(self, **kwargs):
        #
        # Required Args
        try:
            if not isinstance(kwargs['slug'], str):
                raise ValueError("Slug must be a string")

            self.__slug = kwargs['slug']
        except KeyError as key_error:
            raise ValueError("Slug is a required argument") from key_error

        #XXX better validation here
        try:
            if not isinstance(kwargs['memory'], int) or kwargs['memory'] < 8:
                raise ValueError("Memory must be an integer >= 8")

            self.__memory = kwargs['memory']
        except KeyError as key_error:
            raise ValueError("Memory is a required argument") from key_error

        try:
            if not isinstance(kwargs['vcpus'], int) or kwargs['memory'] <1:
                raise ValueError("vcpus must be an integer > 0")

            self.__vcpus = kwargs['vcpus']
        except KeyError as key_error:
            raise ValueError("Vcpus is a required argument") from key_error


        try:
            if not isinstance(kwargs['disk'], int) or kwargs['disk'] < 0:
                raise ValueError(f"disk must be a positive integer: {kwargs['disk']}")

            self.__disk = kwargs['disk']
        except KeyError as key_error:
            raise ValueError("Disk is a required argument") from key_error

        try:
            if not isinstance(kwargs['transfer'], float) or kwargs['transfer'] < 0:
                raise ValueError(f"transfer must a float >= 1: {kwargs['transfer']}")
            self.__transfer = kwargs['transfer']
        except KeyError as key_error:
            raise ValueError("Transfer is a required argument") from key_error

        try:
            if kwargs['price_monthly'] < 0:
                raise ValueError("price_monthly must be a float > 0")

            self.__price_monthly = kwargs['price_monthly']
        except KeyError as key_error:
            raise ValueError("price_monthly is a required argument") from key_error

        try:
            if not isinstance(kwargs['price_hourly'], float) or kwargs['price_hourly'] < 0:
                raise ValueError("price_hourly must be a float > 0")

            self.__price_hourly = kwargs['price_hourly']
        except KeyError as key_error:
            raise ValueError("price_hourly is a required argument") from key_error

        try:
            validate_list_of_strings(kwargs['regions'])
        except KeyError as key_error:
            raise ValueError("Region is a required argument") from key_error
        except ValueError as value_error:
            raise ValueError("Regions must be a list of strings") from value_error

        self.__regions = kwargs['regions']

        try:
            if not isinstance(kwargs['available'], bool):
                raise ValueError("available must be a boolean")

            self.__available = kwargs['available']
        except KeyError as key_error:
            raise ValueError("avaiilable is a requied argument") from key_error

        try:
            if not isinstance(kwargs['description'], str):
                raise ValueError("description must be a string")

            self.__description = kwargs['description']
        except KeyError as key_error:
            raise ValueError("description is a required argument") from key_error


    def slug(self):
        """Size Slug Accessor

            Returns:
                slug (str): Size Slug
        """
        return self.__slug

    def memory(self):
        """Size Memory Accessor

            Returns:
                memory (int): Size Memory
        """
        return self.__memory

    def vcpus(self):
        """Size VCPUs Accessor

            Returns:
                vcpus (int): Size VCPUs
        """
        return self.__vcpus

    def disk(self):
        """Size Disk Accessor

            Returns:
                disk (int): Size Disk
        """
        return self.__disk

    def transfer(self):
        """Size Transfer

            Returns:
                transfer (float): Size float
        """
        return self.__transfer

    def price_monthly(self):
        """Size Price Montly Accessor

            Returns:
                price_montly (float): Size Price Monthly
        """
        return self.__price_monthly

    def price_hourly(self):
        """Size Price Hourly Accessor

            Returns:
                price_hourly (float): Size price hourly
        """
        return self.__price_hourly

    def regions(self):
        """Size Regions Accessor

            Returns:
                regions (:type: list of str): List of regions for size
        """
        return self.__regions

    def available(self):
        """Size Available Accessor

            Returns:
                available (bool): Size availableA
        """
        return self.__available

    def description(self):
        """Size Description Accessor

            Returns:
                description (str): Size description
        """
        return self.__description

    @classmethod
    def list_sizes(cls):
        """List Sizes

            Returns:
                sizes (:type: list of :obj: Size): List of sizes

            Raises:
                ValueError - Invalid Size
                NotAuthorized - Not Authorized to list sizes
                TooManyRequests - Too many requests made
                InternalError - Internal Error
        """
        ret = send_request("GET", SIZES_URL)

        sizes = []
        for size in ret['sizes']:
            sizes.append(Size(**size))

        return sizes



"""
IPv4 Network Object
"""
class V4:
    """V4 Network Object

        Attributes:
            ip_address (str): The IPv4 network interface
            netmask (str): The netmask of the IPv4 network interface
            gateway (str): The gateway of the specified IPv4 network interface.
                For private interfaces, a gateway is not provided.
            type (str): The type of IPv4 network interface (public, private)
    """
    def __init__(self, ip_address, netmask, type, gateway=None):
        #XXX need validation
        self.__ip_address = ip_address
        self.__netmask = netmask
        self.__type = type
        self.__gateway = gateway

    def ip_address(self):
        """V4 IP Address Accessor

            Returns:
                ip_address (str): IPv4 Address
        """
        return self.__ip_address

    def netmask(self):
        """V4 IP Netmask Accessor

            Returns:
                netmask (str): IPv4 Netmask
        """
        return self.__netmask

    def type(self):
        """V4 Type

            Returns:
                type (str): IPv4 type (public|private)
        """
        return self.__type


    def gateway(self):
        """V4 Gateway

            Returns:
                gateway (str): IPv4 gateway
        """
        return self.__gateway



"""
IPv6 Network Object
"""

class V6:
    """Network V6 Object

        Attributes:
            ip_address (str): The IP address of the IPv6 network interface.
            netmask (int): The netmask of the IPv6 network interface.
            gateway (str): The gateway of the specified IPv6 network interface.
            type (str): The type of IPv6 network interface.
                Currently IPv6 private networking is not supported.
    """
    def __init__(self, ip_address, netmask, type, gateway=None):
        self.__ip_address = ip_address
        self.__netmask = netmask
        self.__type = type
        self.__gateway = gateway

    def ip_address(self):
        """V6 IP Address Accessor

            Returns:
                ip_address (str): V6 IP Address
        """
        return self.__ip_address

    def netmask(self):
        """V6 Netmask Accessor

            Return:
                netmask (int): V6 Netmask
        """
        return self.__netmask

    def type(self):
        """V6 Type

            Return:
                type (str): V6 Type
        """
        return self.__type


    def gateway(self):
        """V6 Gateway

            Return:
                gateway (str): V6 Gateway
        """
        return self.__gateway


#XXX Need validation



"""
Common Utility Functions
"""
import re

def validate_list_of_strings(strings):
    """Validates a list is a list of strings

        Args:
            strings (:type: list of str): List of strings

        Raises:
            ValueError - One element of list is not a string
    """
    for string in strings:
        if not isinstance(string, str):
            raise ValueError

def validate_list_of_ints(ints):
    """Validates a list is a list of ints

        Args:
            ints (:type: list of int): List of ints

        Raises:
            ValueError - One element of list is not an integer
    """
    for i in ints:
        if not isinstance(i, int):
            raise ValueError

def validate_date_time(date_time):
    """Validate Date and Time is ISO8601 Format

        Args:
            date_time (str): Date and Time string

        Raises:
            ValueError: String was not of ISO8601 Format
    """
    regex = r'^(-?(?:[1-9][0-9]*)?[0-9]{4})-(1[0-2]|0[1-9])' \
            '-(3[01]|0[1-9]|[12][0-9])T(2[0-3]|[01][0-9]):'\
            '([0-5][0-9]):([0-5][0-9])(\.[0-9]+)?'\
            '(Z|[+-](?:2[0-3]|[01][0-9]):[0-5][0-9])?$'

    date_time_match = re.compile(regex).match
    if date_time_match(date_time) is None:
        raise ValueError



import setuptools

#
# XXX pull from a seperate file like usage
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="py-digitalocean",
    version="0.1.0",
    author="Alan Chester",
    author_email="amcheste@gmail.com",
    url="https://gitlab.com/camphotos/py-digitalocean",
    description="Digital Ocean Python Library",
    long_description="Coming Soon...",
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['requests']
)
